@@ -3,6 +3,7 @@ import pytest
 import os
 from collections import UserDict
 from textwrap import dedent
+import json
 
 
 def example_save_function(self, key, value):
@@ -23,7 +24,7 @@ def dummy_save_function(self, *args, **kwargs):
     pass
 
 
-class Test_Core_Functionality:
+class Test_Initialisation:
     def test_creation_using_existing_dict(self):
         """ CleverDicts can be creates from existing dictionaries """
         x = CleverDict({"total": 6, "usergroup": "Knights of Ni"})
@@ -62,126 +63,27 @@ class Test_Core_Functionality:
         assert x["subject"] == "Python"
 
     def test_creation_using_fromkeys(self):
-        """ Works for initiating with the same value"""
+        """ Works for initiating >1 keys with the same value"""
         x = CleverDict().fromkeys(["a", 1, "what?"], "val")
         assert x.a == "val"
         assert x._1 == "val"
         assert x.what_ == "val"
 
-    def test_raises_error(self):
-        """
-        Attribute and Key errors should be raised as with normal objects/dicts
-        """
-        x = CleverDict()
-        with pytest.raises(AttributeError):
-            x.a
-        with pytest.raises(KeyError):
-            x["a"]
+    def test_creation_using_fromlist(self):
+        """ Works for initiating using a list of iterables """
+        x = CleverDict().fromlist([(1, "one"), [2, "two"], {3, "three"}])
+        assert x._1 == "one"
+        assert x._2 == "two"
+        assert x._3 == "three"
 
-    def test_Expand(self):
-        x = CleverDict({1: "First Entry", " ": "space", "??": "question"})
-        assert x._1 == "First Entry"
-        assert x["_1"] == "First Entry"
-        assert x[1] == "First Entry"
-        with pytest.raises(KeyError):
-            x["1"] = 5
-        with pytest.raises(KeyError):
-            x = CleverDict({1: "First Entry", "1": "space", "??": "question"})
-        x["else"] = "is else"
-        assert x["else"] == "is else"
-        assert x["_else"] == "is else"
-        assert x._else == "is else"
-        with pytest.raises(KeyError):
-            x["?else"] = "other"
-        x._4 = "abc"
-        with pytest.raises(KeyError):
-            x["4"] = "def"
-        x[12345.0] = "klm"
-        assert x._12345 == "klm"
-        x[2.0] = "two-point-0"
-        assert x._2 == "two-point-0"
-        x["11a23bccà~£#@q123b/=€впВМвапрй"] = "abc"
-        assert x._11a23bccà____q123b___впВМвапрй == "abc"
-        assert x["11a23bccà~£#@q123b/=€впВМвапрй"] == "abc"
-        assert x["_11a23bccà____q123b___впВМвапрй"] == "abc"
-
-    def test_data_attribute(self):
-        x = CleverDict()
-        x["data"] = "data"
-        assert x["data"] == "data"
-        assert x.data == "data"
-
-    def test_normalise_unicode(self):
+class Test_Core_Features:
+    def test_tolist(self):
         """
-        Most unicode letters are valid in attribute names
+        Create a list of (k,v) tuples e.g. for serialising dictionaries
+        with numeric keys (which isn't supported in json.dumps/.loads)
         """
-        x = CleverDict({"ветчина_и_яйца$a": "ham and eggs"})
-        x.ве = "be"
-        x["1ве"] = "1be"
-        assert x.ветчина_и_яйца_a == "ham and eggs"
-        assert x.ве == "be"
-        assert x._1ве == "1be"
-
-    def test_value_change(self):
-        """ New attribute values should update dictionary keys & vice versa """
-        x = CleverDict()
-        x.life = 42
-        x["life"] = 43
-        assert x.life == 43
-        assert x["life"] == 43
-        x.life = 42
-        assert x.life == 42
-        assert x["life"] == 42
-        x["1"] = 10
-        assert x["1"] == 10
-        assert x["_1"] == 10
-        assert x._1 == 10
-        x["_1"] = 11
-        assert x["1"] == 11
-        assert x["_1"] == 11
-        assert x._1 == 11
-        x._1 = 12
-        assert x["1"] == 12
-        assert x["_1"] == 12
-        assert x._1 == 12
-        # can't double assign
-        with pytest.raises(KeyError):
-            x["+1"] = 1
-
-    def test_True_False_None_functionality(self):
-        """
-        When setting dictionary keys in Python:
-        d[0] is d[0.0] is d[False]
-        d[1] is d[1.0] is d[True]
-        d[1234.0] is d[1234]
-        If the same key is set in different ways e.g.
-            d = {0: "Zero", False: "Untrue"}
-        the last (rightmost) value overwrites any previous values, so
-            d[0] == "Untrue"
-        Furthermore, keywords like True and False can't be used as attributes.
-        """
-        x = CleverDict()
-        x[0] = 0
-        x[False] = 1
-        x[1] = 2
-        x[True] = 3
-        x[None] = "nothing"
-        assert x[0] == 1
-        assert x[False] == 1
-        assert x["_0"] == 1
-        assert x["_False"] == 1
-        assert x._0 == 1
-        assert x._False == 1
-        assert x[1] == 3
-        assert x[True] == 3
-        assert x["_1"] == 3
-        assert x["_True"] == 3
-        assert x._1 == 3
-        assert x._True == 3
-        assert x[None] == "nothing"
-        assert x["_None"] == "nothing"
-        with pytest.raises(KeyError):
-            x["None"]
+        x = CleverDict({1: "one", 2: "two"})
+        assert x.tolist() == [(1, 'one'), (2, 'two')]
 
     def test_use_as_dict(self):
         d = dict.fromkeys((0, 1, 23, "what?", "a"), "test")
@@ -189,62 +91,6 @@ class Test_Core_Functionality:
         x.setattr_direct("b", 2)
         assert dict(x) == d
 
-    def test_repr_and_eq(self):
-        """
-        Tests that the output from __repr__ can be used to reconstruct the
-        CleverDict object, and __eq__ can be used to compare CleverDict objects."""
-        x = CleverDict()
-        x[0] = 0
-        x[False] = 1
-        x[1] = 2
-        x[True] = 3
-        x.a = 4
-        x["what?"] = 5
-        x.add_alias("a", "c")
-        y = eval(repr(x))
-        assert x == y
-        y.b = 6
-        assert x != y
-        x = CleverDict()
-        assert eval(repr(x)) == x
-        with Expand(False):
-            x = CleverDict({True: 1})
-            assert len(x.get_aliases()) == 1
-            assert CleverDict(eval(repr(x))) == x
-            # check whether _expand has been properly reset
-            x = CleverDict({True: 1})
-            assert len(x.get_aliases()) == 1
-        # empty dict with one variable
-        x = CleverDict()
-        x.setattr_direct("a", 1)
-        assert len(x.get_aliases()) == 0
-        assert eval(repr(x)) == x
-
-    def testupdate(self):
-        x = CleverDict.fromkeys((0, 1, 2, "a", "what?", "return"), 0)
-        y = CleverDict({0: 2, "c": 3})
-        x.update(y)
-        assert x == CleverDict({0: 2, 1: 0, 2: 0, "a": 0, "what?": 0, "return": 0, "c": 3})
-
-    def test_del(self):
-        x = CleverDict()
-        x[1] = 1
-        del x[1]
-        with pytest.raises(KeyError):
-            x[1]
-        x[1] = 1
-        del x["_1"]
-        with pytest.raises(KeyError):
-            x[1]
-        x[1] = 1
-        del x._1
-        with pytest.raises(KeyError):
-            x[1]
-        with pytest.raises(KeyError):
-            del x[1]
-        with pytest.raises(AttributeError):
-            del x._1
-
     def test_add_alias_delete_alias(self):
         """
         Aliases are created automatically after expanding 1, True
@@ -310,6 +156,207 @@ class Test_Core_Functionality:
             x.delete_alias(True)
         assert x.get_aliases("2") == ["2", "_2", "_1", "_True"]
 
+    def test_value_change(self):
+        """ New attribute values should update dictionary keys & vice versa """
+        x = CleverDict()
+        x.life = 42
+        x["life"] = 43
+        assert x.life == 43
+        assert x["life"] == 43
+        x.life = 42
+        assert x.life == 42
+        assert x["life"] == 42
+        x["1"] = 10
+        assert x["1"] == 10
+        assert x["_1"] == 10
+        assert x._1 == 10
+        x["_1"] = 11
+        assert x["1"] == 11
+        assert x["_1"] == 11
+        assert x._1 == 11
+        x._1 = 12
+        assert x["1"] == 12
+        assert x["_1"] == 12
+        assert x._1 == 12
+        # can't double assign
+        with pytest.raises(KeyError):
+            x["+1"] = 1
+
+    def test_info(self, capsys):
+        global c  # globals are not 'seen' by info()
+        z = b = a = c = CleverDict.fromkeys((0, 1, 2, "a", "what?", "return"), 0)
+        c.setattr_direct("b", "B")
+        c.info()  # this prints to stdout
+        out, err = capsys.readouterr()
+        assert out == dedent(
+            """\
+CleverDict:
+    a is b is z
+    a[0] == a['_0'] == a['_False'] == a._0 == a._False == 0
+    a[1] == a['_1'] == a['_True'] == a._1 == a._True == 0
+    a[2] == a['_2'] == a._2 == 0
+    a['a'] == a.a == 0
+    a['what?'] == a['what_'] == a.what_ == 0
+    a['return'] == a['_return'] == a._return == 0
+    a.b == 'B'
+"""
+        )
+        z = b = a = c = CleverDict.fromkeys(["a"], "A")
+        assert c.info(as_str=True) == "CleverDict:\n    a is b is z\n    a['a'] == a.a == 'A'"
+        del a
+        assert c.info(as_str=True) == "CleverDict:\n    b is z\n    b['a'] == b.a == 'A'"
+        del b
+        assert c.info(as_str=True) == "CleverDict:\n    z['a'] == z.a == 'A'"
+        del z
+        assert c.info(as_str=True) == "CleverDict:\n    x['a'] == x.a == 'A'"
+
+
+class Test_Internal_Logic:
+    def test_raises_error(self):
+        """
+        Attribute and Key errors should be raised as with normal objects/dicts
+        """
+        x = CleverDict()
+        with pytest.raises(AttributeError):
+            x.a
+        with pytest.raises(KeyError):
+            x["a"]
+
+    def test_conversions(self):
+        x = CleverDict({1: "First Entry", " ": "space", "??": "question"})
+        assert x._1 == "First Entry"
+        assert x["_1"] == "First Entry"
+        assert x[1] == "First Entry"
+        with pytest.raises(KeyError):
+            x["1"] = 5
+        with pytest.raises(KeyError):
+            x = CleverDict({1: "First Entry", "1": "space", "??": "question"})
+        x["else"] = "is else"
+        assert x["else"] == "is else"
+        assert x["_else"] == "is else"
+        assert x._else == "is else"
+        with pytest.raises(KeyError):
+            x["?else"] = "other"
+        x._4 = "abc"
+        with pytest.raises(KeyError):
+            x["4"] = "def"
+        x[12345.0] = "klm"
+        assert x._12345 == "klm"
+        x[2.0] = "two-point-0"
+        assert x._2 == "two-point-0"
+        x["11a23bccà~£#@q123b/=€впВМвапрй"] = "abc"
+        assert x._11a23bccà____q123b___впВМвапрй == "abc"
+        assert x["11a23bccà~£#@q123b/=€впВМвапрй"] == "abc"
+        assert x["_11a23bccà____q123b___впВМвапрй"] == "abc"
+
+    def test_data_attribute(self):
+        x = CleverDict()
+        x["data"] = "data"
+        assert x["data"] == "data"
+        assert x.data == "data"
+
+    def test_normalise_unicode(self):
+        """
+        Most unicode letters are valid in attribute names
+        """
+        x = CleverDict({"ветчина_и_яйца$a": "ham and eggs"})
+        x.ве = "be"
+        x["1ве"] = "1be"
+        assert x.ветчина_и_яйца_a == "ham and eggs"
+        assert x.ве == "be"
+        assert x._1ве == "1be"
+
+    def test_True_False_None_functionality(self):
+        """
+        When setting dictionary keys in Python:
+        d[0] is d[0.0] is d[False]
+        d[1] is d[1.0] is d[True]
+        d[1234.0] is d[1234]
+        If the same key is set in different ways e.g.
+            d = {0: "Zero", False: "Untrue"}
+        the last (rightmost) value overwrites any previous values, so
+            d[0] == "Untrue"
+        Furthermore, keywords like True and False can't be used as attributes.
+        """
+        x = CleverDict()
+        x[0] = 0
+        x[False] = 1
+        x[1] = 2
+        x[True] = 3
+        x[None] = "nothing"
+        assert x[0] == 1
+        assert x[False] == 1
+        assert x["_0"] == 1
+        assert x["_False"] == 1
+        assert x._0 == 1
+        assert x._False == 1
+        assert x[1] == 3
+        assert x[True] == 3
+        assert x["_1"] == 3
+        assert x["_True"] == 3
+        assert x._1 == 3
+        assert x._True == 3
+        assert x[None] == "nothing"
+        assert x["_None"] == "nothing"
+        with pytest.raises(KeyError):
+            x["None"]
+
+    def test_repr_and_eq(self):
+        """
+        Tests that the output from __repr__ can be used to reconstruct the
+        CleverDict object, and __eq__ can be used to compare CleverDict objects."""
+        x = CleverDict()
+        x[0] = 0
+        x[False] = 1
+        x[1] = 2
+        x[True] = 3
+        x.a = 4
+        x["what?"] = 5
+        x.add_alias("a", "c")
+        y = eval(repr(x))
+        assert x == y
+        y.b = 6
+        assert x != y
+        x = CleverDict()
+        assert eval(repr(x)) == x
+        with Expand(False):
+            x = CleverDict({True: 1})
+            assert len(x.get_aliases()) == 1
+            assert CleverDict(eval(repr(x))) == x
+            # check whether _expand has been properly reset
+            x = CleverDict({True: 1})
+            assert len(x.get_aliases()) == 1
+        # empty dict with one variable
+        x = CleverDict()
+        x.setattr_direct("a", 1)
+        assert len(x.get_aliases()) == 0
+        assert eval(repr(x)) == x
+
+    def testupdate(self):
+        x = CleverDict.fromkeys((0, 1, 2, "a", "what?", "return"), 0)
+        y = CleverDict({0: 2, "c": 3})
+        x.update(y)
+        assert x == CleverDict({0: 2, 1: 0, 2: 0, "a": 0, "what?": 0, "return": 0, "c": 3})
+
+    def test_del(self):
+        x = CleverDict()
+        x[1] = 1
+        del x[1]
+        with pytest.raises(KeyError):
+            x[1]
+        x[1] = 1
+        del x["_1"]
+        with pytest.raises(KeyError):
+            x[1]
+        x[1] = 1
+        del x._1
+        with pytest.raises(KeyError):
+            x[1]
+        with pytest.raises(KeyError):
+            del x[1]
+        with pytest.raises(AttributeError):
+            del x._1
+
     def testget_key(self):
         x = CleverDict.fromkeys(("a", 0, 1, "what?"), 1)
         x.add_alias(0, "zero")
@@ -391,34 +438,6 @@ class Test_Core_Functionality:
             x.get_key("a")
         assert x.get_aliases() == []
 
-    def test_info(self, capsys):
-        global c  # globals are not 'seen' by info()
-        z = b = a = c = CleverDict.fromkeys((0, 1, 2, "a", "what?", "return"), 0)
-        c.setattr_direct("b", "B")
-        c.info()  # this prints to stdout
-        out, err = capsys.readouterr()
-        assert out == dedent(
-            """\
-CleverDict:
-    a is b is z
-    a[0] == a['_0'] == a['_False'] == a._0 == a._False == 0
-    a[1] == a['_1'] == a['_True'] == a._1 == a._True == 0
-    a[2] == a['_2'] == a._2 == 0
-    a['a'] == a.a == 0
-    a['what?'] == a['what_'] == a.what_ == 0
-    a['return'] == a['_return'] == a._return == 0
-    a.b == 'B'
-"""
-        )
-        z = b = a = c = CleverDict.fromkeys(["a"], "A")
-        assert c.info(as_str=True) == "CleverDict:\n    a is b is z\n    a['a'] == a.a == 'A'"
-        del a
-        assert c.info(as_str=True) == "CleverDict:\n    b is z\n    b['a'] == b.a == 'A'"
-        del b
-        assert c.info(as_str=True) == "CleverDict:\n    z['a'] == z.a == 'A'"
-        del z
-        assert c.info(as_str=True) == "CleverDict:\n    x['a'] == x.a == 'A'"
-
     def test_version(self):
         assert isinstance(__version__, str)
 
